<?php 
/**
* Catviz Barchart Library
*
* Copyright (C) 2004 Joost Horward
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License 
* version 2 as published by the Free Software Foundation.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* General Public License for more details.
*
* You should have received a copy of the GNU General Public License 
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330,
* Boston, MA 02111-1307 USA
* 
*
* Generic Barchart library
*
* This barchart package was created for Catviz but is easily usable without it.
* No external dependencies (other than the obvious PHP with GD)
*
* Usage in a nutshell:
* $chart= new TBarChart();
* $chart->setXlabels(array(...));
* $chart->setData(array(...));
* $chart->setTitles(Title,X axis, Y axis);
* $chart->generatePNG();
* Various methods are available to change size, color etc.

*
* @package barchart
* @subpackage barchart
* @author Joost Horward <joost@catviz.org>
* $Date: 2004/10/06 21:41:29 $
* $Author: joosth $
* $Log: barchart.php,v $
* Revision 1.2  2004/10/06 21:41:29  joosth
* Removed TTF support experiment.
*
* Revision 1.1  2004/09/15 17:27:35  joosth
* Added barchart library
*
* Revision$
*
* $Source: /cvsroot/catviz/catviz/lib/barchart.php,v $
*
*
* TODO
* Themes (color/font combinations)
* Presets (day, week, month,year)
* Support both indexed and non-indexed input arrays
* Horizontal values and x-axis legend
*
* Maybe:
* TTF support?
* Nicer looking bars (gradient, shadow, ...)
* Axis thickness choosing
* Background image loading
*
*/

/**
* Creates a max. 4 digit string representing the number.
* Conventient for Y-axis and value labels. They will never be larger than 4 wide.
* Values over 10000 are shown as k an M values, e.g. 76k1, 170M, etc.
* @param string $number The number to be converted
* @param string $ralign Right-align flag, if set value is left-padded with spaces.
*/

function fdstring($number,$ralign=0) {
	$thestring=$number;
	if (strlen($number)<=4)
		$thestring=substr($number,0,4);
	else
		if ($number<10000)
			$thestring=substr($number,0,4);
	
	if (($number>=10000) and ($number<1000000))
		{ $thousands=floor($number / 1000);
		  $units=sprintf('%3d',floor($number % 1000));
		  $thestring=substr("{$thousands}k{$units}",0,4);
		}
	if (($number>=1000000) and ($number<1000000000))
		{ $millions=floor($number / 1000000);
		  $thousands=sprintf('%3d',floor($number % 1000000));
		  $thestring=substr("{$millions}M{$thousands}",0,4);
		}		
	$thestring=trim($thestring);
	if ($thestring!='')
		if ($ralign)
			return sprintf('%4s',$thestring);
		else
			return $thestring;
	else
		return '****';
}
/**
*
* A Barchart generator class
*
* Generates a barchart.
* See doc/barchart.svg and doc/barchart.png for details
*
*
* @author Joost Horward
* @access public
* @package catviz
*/
class TBarChart {
	var $_xLabels; 

	var $_barWidth;
	var $_barSpacing;
	
	var $_yTickSize;
	
	var $_x0;
	var $_y0;
	var $_xSize;
	var $_ySize;
	
	var $_tmargin;
	var $_bmargin;
	var $_lmargin;
	var $_rmargin;
	
	var $_chartWidth;
	var $_chartHeight;
	
	var $_title;
	var $_xtitle;
	var $_ytitle;
	
	var $_im;
	var $_scale;
	
	var $_RGB;
	var $_color;
	
	/**
	* Construct the barchart object
	*
	* Simply sets some default values, nothing is generated/created yet
	*
	*/
	function TBarChart()
	{
		$this->setTitles();
		
		$this->setMargins();
		$this->setBarWidthSpacing();
		$this->setChartHeight();
		
		$this->setRGB('grid',200,200,200);
		$this->setRGB('bar',128,128,200);
		$this->setRGB('image_bg',230,230,230);
		$this->setRGB('chart_bg',255,255,255);
		
		$this->setTextColor();		
		$this->setRGB('axis',0,0,0);
		
	}
	/**
	* set the RGB value of a chart element
	* They are just stored at this point, no colro indices are allocated yet
	*
	* @param string name The name of the chart element.
	* Chart element names are: value_text, xlabel_text, ylabel_text, title_text, xtitle_text, ytitle_text
	*                          grid, bar, image_bg, chart_bg
	* @param int $red The red value (0..255)
	* @param int $green The green value (0..255)
	* @param int $blue The blue value (0..255)
	*/		
	function setRGB($name,$red,$green,$blue){
		$this->_RGB[$name]=array($red,$green,$blue);
	}
	
	/**
	* Allocate color indices in the $this->_im image
	*/	
	function setColorIndexes(){
		foreach ($this->_RGB as $name =>$colors){
			list($red,$green,$blue)=$colors;
			$this->_color[$name]=ImageColorAllocate($this->_im,$red,$green,$blue);
		}		
	}
	
	/**
	* Set all text colors to the same value.
	* Default is black
	* @param int $red The red value (0..255)
	* @param int $green The green value (0..255)
	* @param int $blue The blue value (0..255)	
	*/
	function setTextColor($red=0,$green=0,$blue=0){
		$this->setRGB('value_text' ,$red,$green,$blue);
		$this->setRGB('xlabel_text',$red,$green,$blue);
		$this->setRGB('ylabel_text',$red,$green,$blue);
		$this->setRGB('title_text' ,$red,$green,$blue);
		$this->setRGB('xtitle_text',$red,$green,$blue);
		$this->setRGB('ytitle_text',$red,$green,$blue);
	}
	
	/**
	* Paint rectangles for the whole image background and for the chart background
	*/
	function generateBackground(){
		ImageFilledRectangle($this->_im,0,0,$this->_xSize,$this->_ySize,$this->_color['image_bg']);
		ImageFilledRectangle($this->_im,$this->_lmargin,$this->_tmargin,$this->_xSize-$this->_rmargin,$this->_ySize-$this->_bmargin,$this->_color['chart_bg']);
		
	}
	
	/**
	* Set the chart margins, and recalculate $_x0 and $_y0
	* The assumption is that the chart stays the same so the whole graph resizes 
	* when this function is called
	* $_x0 and $_y0 are recalculated
	* @param int $lmargin The left margin
	* @param int $rmargin The right margin
	* @param int $tmargin The top margin
	* @param int $bmargin The bottom margin
	*/
	function setMargins($lmargin=60,$rmargin=20,$tmargin=60,$bmargin=40){
		$this->_lmargin=$lmargin;
		$this->_rmargin=$rmargin;
		$this->_tmargin=$tmargin;
		$this->_bmargin=$bmargin;
		$this->_x0=$lmargin;
		$this->_y0=$this->_chartHeight+$this->_tmargin;		
	}
	
	/**
	* Set the X labels values
	* @param array $labels The labels array. This is treated as an array of strings
	*/
	function setXlabels($labels){
		$this->_xLabels=$labels;
	}
	
	/**
	* Set the chart data
	* @param array $data An array of numbers 0..999.999.999 
	*/	
	function setData($data){
		$this->_data=$data;
	}
	
	/**
	* Set the chart height
	* Recalculates $_y0 and $_yTicksize 
	* 
	*/
	function setChartHeight($height=200){
		$this->_chartHeight=$height;
		$this->_yTickSize=$height/10;
		$this->_y0=$this->_chartHeight+$this->_tmargin;
	}
	
	/**
	* Set the size of the entire image
	* In case you want a specific size for the entire chart this function is your friend.
	* It maximizes the size of the chart within the specified margins, choosing nice round numbers for ticks and bars
	* Then adds the remaining few pixesl to the right and bottom margin
	* @param int $x The desired x size
	* @param int $y The desired y size
	*/
	function setSize($x,$y){
		$rawy=$y-$this->_tmargin-$this->_bmargin;
		$this->_bmargin+=$rawy % 10; // This is MOD
		$this->_chartHeight=$y-$this->_tmargin-$this->_bmargin;
		
		$rawx=$x-$this->_lmargin-$this->_rmargin-$this->_barSpacing-1;
		$this->_barWidth=floor($rawx/count($this->_xLabels))-$this->_barSpacing;
		$this->_rmargin=$x-$this->_lmargin-(($this->_barWidth+$this->_barSpacing)*count($this->_xLabels))-$this->_barSpacing-1;
		
		$this->_x0=$this->_lmargin;
		$this->_y0=$this->_chartHeight+$this->_tmargin;
		
		$this->_yTickSize=$this->_chartHeight / 10;				
	}
	
	/**
	* Set the bar width and spacing
	* @param int $width The bar width (default 10)
	* @param int $spacing The horizontal spacing between bars 
	*/ 	
	function setBarWidthSpacing($width=10,$spacing=2){
		$this->_barWidth=$width;
		$this->_barSpacing=$spacing;
	}	
	
	/**
	* Set the titles of the chart
	* @param string $title The main title of the chart
	* @param string $xtitle The X axis=is title of the chart
	* @param string $ytitle The Y axis=is title of the chart
	*/
	function setTitles($title='Unnamed Bar Chart',$xtitle='X',$ytitle='Y'){
		$this->_title=$title;
		$this->_xtitle=$xtitle;
		$this->_ytitle=$ytitle;
	}
	
	/**
	* Generate the grid
	* Draws the grid on the image. It consists of horizontal lines only
	*/
	function generateGrid(){
		for ($ypos=0;$ypos<=$this->_chartHeight;$ypos+=$this->_yTickSize){
			ImageLine($this->_im, $this->_x0,$this->_y0-$ypos,$this->_x0+$this->_chartWidth,$this->_y0-$ypos,$this->_color['grid']);
		}		
	}
	
	/** Generate the X axis
	* Generate the X axis X label per bar and the X title 
	*/
	function generateXaxis(){
		ImageLine($this->_im,$this->_x0,$this->_y0,$this->_x0+$this->_chartWidth, $this->_y0,$this->_color['axis']);
		
		/* labels on the X axis */
		$xpos=$this->_x0+$this->_barSpacing;
		$ypos=$this->_y0+2;
		foreach ($this->_xLabels as $xlabel) {
			$stringheight=imagefontheight(1);
			$stringwidth=imagefontwidth(1)*strlen($xlabel);
			$offset=($this->_barWidth-$stringheight)/2;
	
			ImageStringUp($this->_im,1,$xpos+$offset,$ypos+$stringwidth,$xlabel,$this->_color['xlabel_text']);
			$xpos=$xpos+$this->_barWidth+$this->_barSpacing;
		}		
		
		// Print the X axis text
		$stringheight=imagefontheight(2);		
		$stringwidth=imagefontwidth(2)*strlen($this->_xtitle);
		$xpos=$this->_lmargin+($this->_chartWidth/2)-($stringwidth/2);
		ImageString($this->_im,2,$xpos,$this->_y0+$stringheight+10,$this->_xtitle,$this->_color['xtitle_text']);
		
		// This is roughly how TTF support would work. Needs correct positioning, font name selection etc.
		//$points=@imagettftext($this->_im,8,0,$xpos,$this->_y0+$stringheight+10,$this->_color['xtitle_text'],'arial',$this->_xtitle);
		//if (!$points)		
		//	ImageString($this->_im,2,$xpos,$this->_y0+$stringheight+10,$this->_xtitle,$this->_color['xtitle_text']);
	}
	/**
	* Generates the Y axis, tick marks, Tick value labels, and Y title
	*/	
	function generateYaxis(){
		/* Generate the Y axis */
		$ytickvalue=$this->_scale/10;
		ImageLine($this->_im, $this->_x0,$this->_y0,$this->_x0,$this->_tmargin-2,$this->_color['axis']);
		/* generate the Y axis label ticks */
		$yvalue=0;		
		for ($ypos=0;$ypos<=$this->_chartHeight;$ypos+=$this->_yTickSize){
			ImageSetPixel($this->_im, $this->_x0-1,$this->_y0-$ypos,$black);
			$ystring=fdstring($yvalue);
			$stringwidth=imagefontwidth(1)*strlen($ystring);
			$stringwidth=imagefontwidth(1)*strlen($ystring);
			$stringheight=imagefontheight(1);
			ImageString($this->_im,1, $this->_x0-1-$stringwidth-2,$this->_y0-$ypos-($stringheight/2),$ystring,$this->_color['ylabel_text']);
			$yvalue+=$ytickvalue;
		}
		
		// Print the Y axis text
		$stringheight=imagefontheight(2);
		$stringwidth=imagefontwidth(2)*strlen($this->_ytitle);
		ImageStringUp($this->_im,2,$this->_x0-$stringheight-30,($this->_chartHeight/2)+$this->_tmargin+($stringwidth/2),$this->_ytitle,$this->_color['ytitle_text']);
	}
	/**
	* Generate the main chart title
	*/
	function generateChartTitle(){
		$stringheight=imagefontheight(2);
		$stringwidth=imagefontwidth(2)*strlen($this->_title);
		$xpos=$this->_lmargin+($this->_chartWidth/2)-($stringwidth/2);
		$ypos=$stringheight+5;
		//$points=@imagettftext($this->_im,10,0,$xpos,$ypos,$this->_color['title_text'],'arial',$this->_title);
		//if(!$points)
			ImageString($this->_im,2,$xpos,$ypos,$this->_title,$this->_color['title_text']);
	}
	
	/*
	* Generate the whole PNG
	*/		
	function generatePNG(){
		
		$this->_chartWidth=($this->_barWidth+$this->_barSpacing)*count($this->_data)+$this->_barSpacing+1;
		$this->_xSize=$this->_lmargin+$this->_chartWidth+$this->_rmargin;
		
		$this->_ySize=$this->_tmargin+$this->_chartHeight+$this->_bmargin;
		

		$this->_im=ImageCreate($this->_xSize,$this->_ySize);
		
		$this->setColorIndexes();
		
		$this->generateBackground();
		$this->generateGrid();
				
				
		//Determine max scale
		//$scales=array(1,2,5,10,20,50,100,200,500,1000,2000,5000,10000,20000,50000,100000,200000,500000,1000000,2000000,5000000);
		$maxval=max($this->_data);
		$scales=array(5000000,2000000,1000000,500000,200000,100000,50000,20000,10000,5000,2000,1000,500,200,100,50,20,10,5,2,1);
		foreach ($scales as $scale){
			if ($scale>$maxval) $this->_scale=$scale;
		}
		
		$this->generateXaxis();
		$this->generateYaxis();
		$this->generateChartTitle();
		
		$xpos=$this->_x0+$this->_barSpacing;
		// Generate the bars
		foreach ($this->_data as $value) {
			ImageFilledRectangle($this->_im,$xpos,$this->_y0-$value*$this->_chartHeight/$this->_scale,$xpos+$this->_barWidth-1,$this->_y0,$this->_color['bar']);
			$stringheight=imagefontheight(1);
			$offset=($this->_barWidth-$stringheight)/2;
			// Generate the value string if it's not zero
			if ($value)							
				ImageStringUp($this->_im,1,$xpos+$offset,$this->_y0-$value*$this->_chartHeight/$this->_scale-2,fdstring($value),$this->_color['value_text']);
			
			$xpos=$xpos+$this->_barSpacing+$this->_barWidth;
		}
		
		$this->generateXaxis();
		$this->generateYaxis();
		$this->generateChartTitle();
		
		header('Content-Type: image/png');
		ImagePNG($this->_im);		
	}
}
?>
